# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# * The software is provided "as is" without any warranty, neither express
#   nor implied.
# * In no event will the author be held liable for any damage(s) arising
#   from the use of the software.
# * Redistribution of the software is allowed only in unmodified form.
# * Permission is granted to use the software solely for personal private
#   and non-commercial purposes.
# * An individuals use of the software in his or her capacity or function
#   as an agent, (independent) contractor, employee, member or officer of
#   a business, corporation or organization (commercial or non-commercial)
#   does not qualify as personal private and non-commercial purpose.
# * Without written approval from the author the software must not be used
#   for a business, for commercial, corporate, governmental, military or
#   organizational purposes of any kind, or in a commercial, corporate,
#   governmental, military or organizational environment of any kind.

if [ $# -eq 1 -a "$1" == 'amd64' ]
then
as --64 --no-pad-sections --noexecstack -o string.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# Unix System V calling convention for AMD64 platform:
# - first 6 integer or pointer arguments (from left to right) are passed
#   in registers RDI/R7, RSI/R6, RDX/R2, RCX/R1, R8 and R9
#   (R10 is used as static chain pointer in case of nested functions);
# - surplus arguments are pushed on stack in reverse order (from right to
#   left), 8-byte aligned;
# - 128-bit integer arguments are passed as pair of 64-bit integer arguments,
#   low part before/below high part;
# - 128-bit integer result is returned in registers RAX/R0 (low part) and
#   RDX/R2 (high part);
# - 64-bit integer or pointer result is returned in register RAX/R0;
# - 32-bit integer result is returned in register EAX;
# - registers RBX/R3, RSP/R4, RBP/R5, R12, R13, R14 and R15 must be
#   preserved;
# - registers RAX/R0, RCX/R1, RDX/R2, RSI/R6, RDI/R7, R8, R9, R10 (in
#   case of normal functions) and R11 are volatile and can be clobbered;
# - stack is 16-byte aligned: callee must decrement RSP by 8+n*16 bytes
#   before calling other functions (CALL instruction pushes 8 bytes);
# - a "red zone" of 128 bytes below the stack pointer can be clobbered.

.file	"string.s"
.arch	generic64
.code64
.ident	"Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>"
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strcat.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcat.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of destination
					# rsi = address of source
strcat:
.ifdef VARIANT
	mov	rdx, rdi		# rdx = address of destination
	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	repne	scasb
	dec	rdi			# rdi = end of destination (address of '\0')
	xchg	rdi, rsi		# rsi = end of destination (address of '\0'),
					# rdi = address of source
#	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	repne	scasb
	not	rcx			# rcx = length of source (including '\0')
	sub	rdi, rcx		# rdi = address of source
	xchg	rdi, rsi		# rsi = address of source,
					# rdi = end of destination (address of '\0')
	rep	movsb
	mov	rax, rdx		# rax = address of destination
	ret
.else # VARIANT
	mov	r11, rdi		# r11 = address of destination
	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	mov	rdi, rsi		# rdi = address of source
	repne	scasb
	not	rcx
	mov	rdx, rcx		# rdx = length of source (including '\0')
#	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	mov	rdi, r11		# rdi = address of destination
	repne	scasb
	dec	rdi			# rdi = end of destination (address of '\0')
	mov	rcx, rdx		# rcx = length of source (including '\0')
	rep	movsb
	mov	rax, r11		# rax = address of destination
	ret
.endif # VARIANT
.size	strcat, .-strcat
.type	strcat, @function
.global	strcat
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strcat.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcat.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of destination
					# rsi = address of source
strcat:
	mov	rdx, rdi		# rdx = address of destination
	xor	eax, eax		# rax = '\0'
.Lscan:
	scasb
	jne	.Lscan			# *destination++ <> '\0'?

	dec	rdi			# rdi = end of destination (address of '\0')
.Lcopy:
	lodsb				# al = *source++
	stosb				# *destination++ = al
	test	al, al
	jnz	.Lcopy			# al <> '\0'?

	mov	rax, rdx		# rdx = address of destination
	ret

.size	strcat, .-strcat
.type	strcat, @function
.global	strcat
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strchr.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# sil = character
strchr:
	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	repne	scasb
	not	rcx			# rcx = length of string (including '\0')
	sub	rdi, rcx		# rdi = address of string
	mov	eax, esi		# rax = character
	repne	scasb
	lea	rax, [rdi-1]		# rax = address of character
	test	rcx, rcx
	cmovz	rax, rcx		# rax = (rcx = 0) ? 0 : address of character
	ret

.size	strchr, .-strchr
.type	strchr, @function
.global	strchr
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strchr.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# sil = character
strchr:
	xchg	rdi, rsi		# rsi = address of string,
					# dil = character
.Lloop:
	lodsb				# al = *string++
	cmp	al, dil
	je	.Lmatch			# al = character?

	test	al, al
	jnz	.Lloop			# al <> '\0'?

	xor	eax, eax
	ret
.Lmatch:
	lea	rax, [rsi-1]
	ret

.size	strchr, .-strchr
.type	strchr, @function
.global	strchr
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strcmp.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcmp.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of left string
					# rsi = address of right string
strcmp:
	xor	eax, eax		# rax = 0
	cmp	rdi, rsi
	je	.Lequal			# address of left string = address of right string?

#	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	repne	scasb
	not	rcx			# rcx = length of left string (including '\0')
	sub	rdi, rcx		# rdi = address of left string
#	xor	eax, eax		# rax = 0
	repe	cmpsb
	cmc				# CF = (left string < right string)
	seta	al			# rax = (left string > right string) ? 1 : 0
	sbb	rax, 0			# rax = (left string < right string) ? -1
					#     : (left string > right string) ? 1 : 0
					#     = {-1, 0, 1}
.Lequal:
	ret

.size	strcmp, .-strcmp
.type	strcmp, @function
.global	strcmp
.end
+EOF

# NOTE: strcoll() is another implementation of strcmp()

as --64 --no-pad-sections --noexecstack -o strcoll.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcoll.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of left string
					# rsi = address of right string
strcoll:
	cmp	rdi, rsi
	je	.Lequal			# address of left string = address of right string?
.Lloop:
	lodsb				# al = *right++
	scasb
	jne	.Lnotequal		# al <> *left++?

	test	al, al
	jnz	.Lloop			# al <> '\0'?
.Lequal:
	xor	eax, eax
	ret
.Lnotequal:
	cmc				# CF = (*left < *right)
	sbb	rax, rax		# rax = (*left < *right) ? -1 : 0
	or	rax, 1			# rax = (*left < *right) ? -1
					#     : (*left > *right) ? 1 : 0
					#     = {-1, 0, 1}
	ret

.size	strcoll, .-strcoll
.type	strcoll, @function
.global	strcoll
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strcoll.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcoll.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of left string
					# rsi = address of right string
strcoll:
	sub	rsi, rdi
	jz	.Lequal			# address of right string = address of left string?
.Lloop:
	mov	al, [rdi]		# al = *left
	cmp	al, [rdi+rsi]
	jne	.Lnotequal		# al = *right?

	inc	rdi
	test	al, al
	jnz	.Lloop			# al <> '\0'?
.Lequal:
	xor	eax, eax
	ret
.Lnotequal:
	sbb	rax, rax		# rax = (*left < *right) ? -1 : 0
	or	rax, 1			# rax = (*left < *right) ? -1
					#     : (*left > *right) ? 1 : 0
					#     = {-1, 0, 1}
	ret

.size	strcoll, .-strcoll
.type	strcoll, @function
.global	strcoll
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strcpy.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcpy.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of destination
					# rsi = address of source
strcpy:
	mov	rdx, rdi		# rdx = address of destination
	mov	rdi, rsi		# rdi = address of source
	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	repne	scasb
	not	rcx			# rcx = length of source (including '\0')
	sub	rdi, rcx		# rdi = address of source
	mov	rsi, rdi		# rsi = address of source
	mov	rdi, rdx		# rdi = address of destination
	rep	movsb
	mov	rax, rdi		# rax = address of destination
	ret

.size	strcpy, .-strcpy
.type	strcpy, @function
.global	strcpy
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strcpy.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcpy.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of destination
					# rsi = address of source
strcpy:
	mov	rdx, rdi		# rdx = address of destination
.Lloop:
	lodsb				# al = *source++
	stosb				# *destination++ = al
	test	al, al
	jnz	.Lloop			# al <> '\0'?

	mov	rax, rdx		# rdx = address of destination
	ret

.size	strcpy, .-strcpy
.type	strcpy, @function
.global	strcpy
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strcspn.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcspn.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# rsi = address of delimiter
strcspn:
	xor	eax, eax		# rax = '\0'
	cmp	al, [rsi]
	je	.Lempty			# delimiter[0] = '\0'?

	push	rax
	push	rax
	push	rax
	push	rax			# bitmap[0..256] = 0,
					# rsp = address of bitmap
.Lsetup:
	bts	[rsp], rax		# bitmap[rax] = 1
	lodsb				# rax = *delimiter++
	cmp	al, ah
	jne	.Lsetup			# rax <> '\0'?

	mov	rsi, rdi		# rsi = address of string
.Lskip:
	lodsb				# rax = *string++
	bt	[rsp], rax
	jnc	.Lskip			# bitmap[rax] = 0?

	sbb	rsi, rdi
	mov	rax, rsi		# rax = count of non-matching characters
	add	rsp, 32
	ret
.Lempty:
	mov	rsi, rdi		# rsi = address of string
.Llength:
	scasb
	jne	.Llength		# *string++ <> '\0'?

	stc
	sbb	rdi, rsi
	mov	rax, rdi		# rax = count of characters
	ret

.size	strcspn, .-strcspn
.type	strcspn, @function
.global	strcspn
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strlen.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strlen.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
strlen:
	mov	r8, 0x0101010101010101
.if 0
	mov	r9, 0x8080808080808080
.elseif 0
	imul	r9, r8, 128		# r9 = 0x8080808080808080
.else
	mov	r9, r8
	ror	r9, 1			# r9 = 0x8080808080808080
.endif
	mov	rsi, rdi
	mov	ecx, edi
	and	ecx, 7			# rcx = address of string % 8
					#     = 8 - number of unaligned characters
	jz	.Laligned		# address of string % 8 = 0?
.Lunaligned:
	sub	rdi, rcx		# rdi = aligned address before string
	shl	ecx, 3			# rcx = (8 - number of unaligned characters) * 8
					#     = 64 - number of unaligned bits
.ifdef AMD
	stc
	sbb	rax, rax		# rax = ~0
.else
	or	rax, -1			# rax = ~0
.endif
	shl	rax, cl			# rax = ~0 for unaligned characters, 0 elsewhere
	not	rax			# rax = 0 for unaligned characters, ~0 elsewhere
	or	rax, [rdi]		# rax = unaligned characters
	jmp	.Ltwiddle
.Lnext:
	add	rdi, 8			# rdi = address of next 8 aligned characters
.Laligned:
	mov	rax, [rdi]		# rax = next 8 aligned characters
.Ltwiddle:
	mov	rcx, rax
	sub	rax, r8
	not	rcx
	and	rax, r9
	and	rcx, rax		# rax = '\200' for matching characters, '\0' elsewhere
	jz	.Lnext			# no '\0' in any character?
.Lmatch:
	bsf	rax, rcx		# rax = offset of '\0' * 8 + 7
					#     = {7, 15, 23, 31, 39, 47, 55, 63}
	shr	eax, 3			# rax = offset of '\0'
					#     = {0, 1, 2, 3, 4, 5, 6, 7}
	add	rax, rdi		# rax = address of '\0'
	sub	rax, rsi		# rax = length of string
	ret

.size	strlen, .-strlen
.type	strlen, @function
.global	strlen
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strlen.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strlen.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
strlen:
	mov	rsi, rdi		# rsi = address of string
	xor	eax, eax		# rax = '\0'
.Lloop:
	scasb
	jne	.Lloop			# *string++ <> '\0'?

	stc
	sbb	rdi, rsi
	mov	rax, rdi
	ret

.size	strlen, .-strlen
.type	strlen, @function
.global	strlen
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strlen.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strlen.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
strlen:
	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	repne	scasb
	not	rcx			# rcx = length of string (including '\0')
	dec	rcx
	mov	rax, rcx
	ret

.size	strlen, .-strlen
.type	strlen, @function
.global	strlen
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strncat.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strncat.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of destination
					# rsi = address of source
					# rdx = count
strncat:
	mov	r11, rdi		# r11 = address of destination
	xor	eax, eax		# rax = '\0'
	mov	rcx, rdx		# rcx = count
	mov	rdi, rsi		# rdi = address of source
	repne	scasb
	sub	rdx, rcx		# rdx = length of source (including '\0')
#	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	mov	rdi, r11		# rdi = address of destination
	repne	scasb
	dec	rdi			# rdi = end of destination (address of '\0')
	mov	rcx, rdx		# rcx = length of source (including '\0')
	rep	movsb
#	xor	eax, eax		# rax = '\0'
	stosb
	mov	rax, r11		# rax = address of destination
	ret

.size	strncat, .-strncat
.type	strncat, @function
.global	strncat
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strncmp.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strncmp.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of left string
					# rsi = address of right string
					# rdx = count
strncmp:
	xor	eax, eax		# rax = 0
	cmp	rdi, rsi
	je	.Lequal			# address of left string = address of right string?

	test	rdx, rdx
	jz	.Lequal			# count = 0?

#	xor	eax, eax		# rax = 0,
#					# CF = 0, ZF = 1 (required when rcx is 0)
	mov	rcx, rdx		# rcx = count
	repe	cmpsb
	cmc				# CF = (left string < right string)
	seta	al			# rax = (left string > right string) ? 1 : 0
	sbb	eax, 0			# rax = (left string < right string) ? -1
					#     : (left string > right string) ? 1 : 0
					#     = {-1, 0, 1}
.Lequal:
	ret

.size	strncmp, .-strncmp
.type	strncmp, @function
.global	strncmp
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strncmp.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strncmp.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of left string
					# rsi = address of right string
					# rdx = count
strncmp:
#	cmp	rdi, rsi
#	je	.Lequal			# address of left string = address of right string?

	test	rdx, rdx
	jz	.Lequal			# count = 0?
.Lloop:
	lodsb				# al = *right++
	scasb
	jne	.Lnotequal		# al <> *left++?

	test	al, al
	jz	.Lequal			# al = '\0'?

	dec	rdx
	jnz	.Lloop			# rdx <> 0?
.Lequal:
	xor	eax, eax
	ret
.Lnotequal:
	cmc				# CF = (*left < *right)
	sbb	rax, rax		# rax = (*left < *right) ? -1 : 0
	or	rax, 1			# rax = (*left > *right) ? 1
					#     : (*left < *right) ? -1 : 0
					#     = {1, 0, -1}
	ret

.size	strncmp, .-strncmp
.type	strncmp, @function
.global	strncmp
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strncpy.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strncpy.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of destination
					# rsi = address of source
					# rdx = count
strncpy:
	mov	r11, rdi		# r11 = address of destination
	xor	eax, eax		# rax = '\0'
	mov	rcx, rdx		# rcx = count
	mov	rdi, rsi		# rdi = address of source
	repne	scasb
	sub	rcx, rdx
	neg	rcx			# rcx = length of source (including '\0')
	sub	rdx, rcx		# rdx = count - length of source (including '\0')
	mov	rdi, r11		# rdi = address of destination
	rep	movsb
	mov	rcx, rdx		# rcx = count - length of source (including '\0')
#	xor	eax, eax		# rax = '\0'
	rep	stosb
	mov	rax, r11		# rax = address of destination
	ret

.size	strncpy, .-strncpy
.type	strncpy, @function
.global	strncpy
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strnlen.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strnlen.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# rsi = count
strnlen:
	xor	eax, eax		# rax = '\0'
	mov	rcx, rsi		# rcx = count
	test	rdi, rdi		# ZF = 0 (required when count is 0)
	repne	scasb			# rcx = (length of string < count)
					#     ? count - (length of string + 1) : 0
	neg	rcx			# CF = (rcx <> 0)
					#    = ([rdi] = '\0')
					#    = (length of string < count),
					# rcx = (length of string < count)
					#     ? length of string + 1 - count : 0
	sbb	rcx, rax		# rcx = (length of string < count)
					#     ? length of string - count : 0
	add	rcx, rsi		# rcx = (length of string < count)
					#     ? length of string : count
	mov	rax, rcx
	ret

.size	strnlen, .-strnlen
.type	strnlen, @function
.global	strnlen
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strnlen.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strnlen.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# rsi = count
strnlen:
	test	rsi, rsi
	jz	.Lzero			# count = 0?

	mov	rdx, rdi		# rdx = address of string
	mov	rcx, rsi		# rcx = count
	xor	eax, eax		# rax = '\0'
.Lloop:
	scasb
	je	.Lend			# *string++ = '\0'?

	dec	rcx
	jnz	.Lloop			# rcx <> 0?
.Lzero:
	mov	rax, rsi		# rax = count
	ret
.Lend:
	stc
	sbb	rdi, rdx
	mov	rax, rdi
	ret

.size	strnlen, .-strnlen
.type	strnlen, @function
.global	strnlen
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strnset.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strnset.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# rsi = character
					# rdx = count
strnset:
	mov	r11, rdi		# r11 = address of string
	xor	eax, eax		# rax = '\0'
	mov	rcx, rdx		# rcx = count
	repne	scasb			# rcx = (length of string < count)
					#     ? count - (length of string + 1) : 0
	sete	al			# rax = (length of string < count) ? 1 : 0
	add	rcx, rax		# rcx = (length of string < count)
					#     ? count - length of string : 0
	sub	rdx, rcx		# rdx = (length of string < count)
					#     ? length of string : count
	mov	rcx, rdx
	mov	eax, esi		# rax = character
	mov	rdi, r11		# rdi = address of string
	rep	stosb
	mov	rax, r11		# rax = address of string
	ret

.size	strnset, .-strnset
.type	strnset, @function
.global	strnset
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strpbrk.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strpbrk.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# rsi = address of delimiter
strpbrk:
	xor	eax, eax		# rax = '\0'
	cmp	al, [rsi]
	je	.Lempty			# delimiter[0] = '\0'?

	push	rax
	push	rax
	push	rax
	push	rax			# bitmap[0..256] = 0,
					# rsp = address of bitmap
.Lsetup:
	bts	[rsp], rax		# bitmap[rax] = 1
	lodsb				# rax = *delimiter++
	cmp	al, ah
	jne	.Lsetup			# rax <> '\0'?

	mov	rsi, rdi		# rsi = address of string
.Lloop:
	lodsb				# rax = *string++
	bt	[rsp], rax
	jnc	.Lloop			# bitmap[rax] = 0?

	dec	rsi			# rsi = --string
	neg	al
	sbb	rax, rax		# rax = (*string = '\0') ? 0 : ~0
	and	rax, rsi		# rax = (*string = '\0') ? 0 : current address of string
	add	rsp, 32
.Lempty:
	ret

.size	strpbrk, .-strpbrk
.type	strpbrk, @function
.global	strpbrk
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strrchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strrchr.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# sil = character
strrchr:
	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	repne	scasb
	not	rcx			# rcx = length of string (including '\0')
	dec	rdi			# rdi = end of string (address of '\0')
	mov	eax, esi		# rax = character
	std
	repne	scasb
	cld
	lea	rax, [rdi+1]		# rax = address of character
	test	rcx, rcx
	cmovz	rax, rcx		# rax = (rcx = 0) ? 0 : address of character
	ret

.size	strrchr, .-strrchr
.type	strrchr, @function
.global	strrchr
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strrchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strrchr.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# sil = character
strrchr:
	xchg	rdi, rsi		# rsi = address of string,
					# dil = character
	mov	edx, 1
.Lloop:
	lodsb				# al = *string++
	cmp	al, dil
	cmove	rdx, rsi		# rdx = (al <> character) ? rdx : rsi

	test	al, al
	jnz	.Lloop			# al <> '\0'?

	lea	rax, [rdx-1]
	ret

.size	strrchr, .-strrchr
.type	strrchr, @function
.global	strrchr
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strset.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strset.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# sil = character
strset:
	mov	rdx, rdi		# rdx = address of string
	xor	eax, eax		# rax = '\0'
.ifdef AMD
	stc
	sbb	rcx, rcx		# rcx = -1
.else
	or	rcx, -1			# rcx = -1
.endif
	repne	scasb
	not	rcx			# rcx = length of string (including '\0')
	mov	rdi, rdx		# rdi = address of string
	dec	rcx			# rcx = length of string
	mov	eax, esi		# rax = character
	rep	stosb
	mov	rax, rdx		# rax = address of string
	ret

.size	strset, .-strset
.type	strset, @function
.global	strset
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strspn.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strspn.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# rsi = address of delimiter
strspn:
	xor	eax, eax		# rax = '\0'
	cmp	al, [rsi]
	je	.Lempty			# delimiter[0] = '\0'?

	push	rax
	push	rax
	push	rax
	push	rax			# bitmap[0..256] = 0,
					# rsp = address of bitmap
	lodsb				# rax = *delimiter++
.Lsetup:
	bts	[rsp], rax		# bitmap[rax] = 1
	lodsb				# rax = *delimiter++
	cmp	al, ah
	jne	.Lsetup			# rax <> '\0'?

	mov	rsi, rdi		# rsi = address of string
.Lskip:
	lodsb				# rax = *string++
	bt	[rsp], rax
	jc	.Lskip			# bitmap[rax] = 1?

	stc
	sbb	rsi, rdi
	mov	rax, rsi		# rax = count of matching characters
	add	rsp, 32
.Lempty:
	ret

.size	strspn, .-strspn
.type	strspn, @function
.global	strspn
.end
+EOF

as --64 --no-pad-sections --noexecstack -o strtok_r.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strtok_r.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of string
					# rsi = address of delimiter
					# rdx = address of address of next character
strtok_r:
	xor	eax, eax		# rax = 0 = '\0'

	test	rdi, rdi
	jnz	.Lstart			# address of string <> 0?

	or	rdi, [rdx]		# rdi = address of next character
	jz	.Lnull			# address of next character = 0 = address of string?
.Lstart:
	cmp	al, [rdi]
	je	.Lnull			# string[0] = '\0'?

	cmp	al, [rsi]
	je	.Lempty			# delimiter[0] = '\0'?

	push	rax
	push	rax
	push	rax
	push	rax			# bitmap[0..256] = 0,
					# rsp = address of bitmap
	lodsb				# rax = *delimiter++
.Lsetup:
	bts	[rsp], rax		# bitmap[rax] = 1
	lodsb				# rax = *delimiter++
	cmp	al, ah
	jne	.Lsetup			# rax <> '\0'?

	mov	rsi, rdi		# rsi = address of string
.Lskip:
	lodsb				# rax = *string++
	bt	[rsp], rax
	jc	.Lskip			# bitmap[rax] = 1?
					#  (rax is a delimiter?)
	cmp	al, ah
	je	.Lnone			# rax = '\0'?

	mov	al, ah			# rax = 0
	bts	[rsp], rax		# bitmap['\0'] = 1
	lea	rdi, [rsi-1]		# rdi = address of token
.Ltoken:
	lodsb				# rax = *string++
	bt	[rsp], rax
	jnc	.Ltoken			# bitmap[rax] = 0?
					#  (rax is not a delimiter?)
	cmp	al, ah
	je	.Llast			# rax = '\0'?

	mov	[rsi-1], ah		# string[-1] = '\0'
					#  (terminate token)
	mov	[rdx], rsi		# address of next character = current address of string
	mov	rax, rdi		# rax = address of token
	add	rsp, 32
	ret
.Llast:
	mov	[rdx], rax		# address of next character = 0
	mov	rax, rdi		# rax = address of token
	add	rsp, 32
	ret
.Lnone:
	mov	[rdx], rax		# address of next character = 0
	add	rsp, 32
	ret
.Lempty:
	mov	[rdx], rax		# address of next character = 0
	mov	rax, rdi		# rax = address of token
	ret
.Lnull:
	mov	[rdx], rax		# address of next character = 0
	ret

.size	strtok_r, .-strtok_r
.type	strtok_r, @function
.global	strtok_r
.end
+EOF

ar rsv string-amd64.a string.o strcat.o strchr.o strcmp.o strcoll.o strcpy.o strcspn.o strlen.o strncat.o strncmp.o strncpy.o strnlen.o strnset.o strpbrk.o strrchr.o strset.o strspn.o strtok_r.o
rm string.o strcat.o strchr.o strcmp.o strcoll.o strcpy.o strcspn.o strlen.o strncat.o strncmp.o strncpy.o strnlen.o strnset.o strpbrk.o strrchr.o strset.o strspn.o strtok_r.o
elif [ $# -eq 1 -a "$1" = 'i386' ]
then
as --32 --no-pad-sections --noexecstack -o string.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# "cdecl" calling and naming convention for i386 platform:
# - arguments are pushed on stack in reverse order (from right to left),
#   4-byte aligned;
# - 64-bit integer arguments are passed as pair of 32-bit integer arguments,
#   low part below high part;
# - 64-bit integer result is returned in registers EAX (low part) and
#   EDX (high part);
# - 32-bit integer or pointer result is returned in register EAX;
# - registers EAX, ECX and EDX are volatile and can be clobbered;
# - registers EBX, ESP, EBP, ESI and EDI must be preserved.

.file	"string.s"
.arch	generic32
.code32
.ident	"Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>"
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strcat.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcat.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = address of source
					# [esp+4] = address of destination
strcat:
	mov	edx, edi
	mov	edi, [esp+8]		# edi = address of source string
	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	not	ecx			# ecx = length of source string (including '\0')
	push	ecx
	mov	edi, [esp+8]		# edi = address of destination string
#	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	dec	edi			# edi = address of '\0'
					#     = end of destination string
	mov	eax, esi
	mov	esi, [esp+12]		# esi = address of source string
	pop	ecx			# ecx = length of source string (including '\0')
	rep	movsb
	mov	edi, edx
	mov	esi, eax
	mov	eax, [esp+4]		# eax = address of destination string
	ret

.size	strcat, .-strcat
.type	strcat, @function
.global	strcat
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strcat.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcat.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = address of source
					# [esp+4] = address of destination
strcat:
	mov	edx, [esp+4]		# edx = address of destination
	mov	ecx, [esp+8]		# ecx = address of source
.Lskip:
	cmp	byte ptr [edx], 0
	lea	edx, [edx+1]		# edx = ++destination
	jne	.Lskip

	dec	edx
	sub	ecx, edx
.Lcopy:
	mov	al, [ecx+edx]
	mov	[edx], al
	inc	edx			# edx = ++destination
	test	al, al
	jnz	.Lcopy

	mov	eax, [esp+4]		# eax = address of destination
	ret

.size	strcat, .-strcat
.type	strcat, @function
.global	strcat
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strchr.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = character
					# [esp+4] = address of string
strchr:
	mov	edx, edi
	mov	edi, [esp+4]		# edi = address of string
	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	not	ecx			# ecx = length of string (including '\0')
	sub	edi, ecx		# edi = address of string
	mov	eax, [esp+8]		# eax = character
	repne	scasb
	dec	edi			# edi = address of character
	neg	ecx			# CF = (ecx <> 0)
					#    = ([edi] = character)
	sbb	eax, eax		# eax = (ecx = 0) ? 0 : -1
	and	eax, edi		# eax = (ecx = 0) ? 0 : address of character
	mov	edi, edx
	ret

.size	strchr, .-strchr
.type	strchr, @function
.global	strchr
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strcmp.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcmp.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = address of right string
					# [esp+4] = address of left string
strcmp:
	push	esi
	push	edi
	xor	eax, eax		# eax = 0
	mov	esi, [esp+12]		# esi = address of left string
	mov	edi, [esp+16]		# edi = address of right string
	cmp	edi, esi
	je	.Lequal			# address of right string = address of left string?

#	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	not	ecx			# ecx = length of right string (including '\0')
	sub	edi, ecx		# edi = address of right string
#	xor	eax, eax		# eax = 0
	repe	cmpsb
	seta	al			# eax = (left string > right string) ? 1 : 0
	sbb	eax, 0			# eax = (left string < right string) ? -1
					#     : (left string > right string) ? 1 : 0
					#     = {-1, 0, 1}
.Lequal:
	pop	edi
	pop	esi
	ret

.size	strcmp, .-strcmp
.type	strcmp, @function
.global	strcmp
.end
+EOF

# NOTE: strcoll() is another implementation of strcmp()

as --32 --no-pad-sections --noexecstack -o strcoll.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcoll.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = address of right string
					# [esp+4] = address of left string
strcoll:
	mov	ecx, [esp+4]		# ecx = address of left string
	mov	edx, [esp+8]		# edx = address of right string
	sub	edx, ecx
	jz	.Lequal			# address of right string = address of left string?
.Lcompare:
	mov	al, [ecx]
	cmp	al, [ecx+edx]
	jne	.Lnotequal

	inc	ecx
	test	al, al
	jnz	.Lcompare
.Lequal:
	xor	eax, eax
	ret
.Lnotequal:
	sbb	eax, eax		# eax = (left string < right string) ? -1 : 0
	or	eax, 1			# eax = (left string < right string) ? -1
					#     : (left string > right string) ? 1 : 0
	ret

.size	strcoll, .-strcoll
.type	strcoll, @function
.global	strcoll
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strcpy.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcpy.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = address of source
					# [esp+4] = address of destination
strcpy:
	mov	edx, edi
	mov	edi, [esp+8]		# edi = address of source
	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	not	ecx			# ecx = length of source (including '\0')
	mov	eax, esi
	mov	esi, [esp+8]		# esi = address of source
	mov	edi, [esp+4]		# edi = address of destination
	rep	movsb
	mov	edi, edx
	mov	esi, eax
	mov	eax, [esp+4]		# eax = address of destination
	ret

.size	strcpy, .-strcpy
.type	strcpy, @function
.global	strcpy
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strcpy.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcpy.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = address of source
					# [esp+4] = address of destination
strcpy:
	mov	edx, [esp+4]		# edx = address of destination
	mov	ecx, [esp+8]		# ecx = address of source
	sub	ecx, edx
	jz	.Lsame
.Lcopy:
	mov	al, [ecx+edx]
	mov	[edx], al
	inc	edx			# edx = ++destination
	test	al, al
	jnz	.Lcopy

	mov	eax, [esp+4]		# eax = address of destination
	ret
.Lsame:
	mov	eax, edx		# eax = address of destination
	ret

.size	strcpy, .-strcpy
.type	strcpy, @function
.global	strcpy
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strcspn.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strcspn.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = address of delimiter
					# [esp+4] = address of string
strcspn:
	mov	eax, [esp+4]		# eax = address of string
	mov	edx, [esp+8]		# edx = address of delimiter
	xor	ecx, ecx		# ecx = 0
	cmp	cl, [edx]
	je	.Lempty			# delimiter[0] = '\0'?

	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx			# bitmap[0..255] = 0,
					# esp = address of bitmap
.Lsetup:
	bts	[esp], ecx		# bitmap[ecx] = 1
	mov	cl, [edx]		# ecx = *delimiter
	inc	edx			# edx = ++delimiter
	cmp	cl, ch
	jne	.Lsetup			# ecx <> '\0'?

	mov	edx, eax		# edx = address of string
.Lskip:
	mov	cl, [eax]		# ecx = *string
	inc	eax			# eax = ++string
	bt	[esp], ecx
	jnc	.Lskip			# bitmap[ecx] = 0 (no match)?
.Lstop:
	sbb	eax, edx		# eax = count of non-matching characters
	add	esp, 32
	ret
.Lempty:
	mov	edx, eax		# edx = address of string
.Llength:
	inc	eax			# eax = ++string
	cmp	cl, [eax-1]
	jne	.Llength
.if 1
	dec	eax			# eax = --string
	sub	eax, edx
.else
	stc
	sbb	eax, edx		# eax = count of characters
.endif
	ret

.size	strcspn, .-strcspn
.type	strcspn, @function
.global	strcspn
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strlen.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strlen.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+4] = address of string
strlen:
	mov	edx, edi
	mov	edi, [esp+4]		# edi = address of string
	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
.if 1
	not	ecx			# ecx = length of string (including '\0')
	dec	ecx
	mov	eax, ecx
.else
	mov	eax, -2
	sub	eax, ecx
.endif
	mov	edi, edx
	ret

.size	strlen, .-strlen
.type	strlen, @function
.global	strlen
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strlen.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strlen.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+4] = address of string
strlen:
	mov	eax, [esp+4]		# eax = address of string
	mov	edx, eax
.Llength:
	inc	eax			# eax = ++string
	cmp	byte ptr [eax-1], 0
	jne	.Llength

	dec	eax
	sub	eax, edx
	ret

.size	strlen, .-strlen
.type	strlen, @function
.global	strlen
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strlen.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strlen.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+4] = address of string
strlen:
	mov	edx, [esp+4]		# edx = address of string
	mov	ecx, edx
	and	ecx, 3			# ecx = address of string % 4
					#     = 4 - number of unaligned characters
	jz	.Laligned		# address of string % 4 = 0?
.Lunaligned:
	sub	edx, ecx		# edx = aligned address before string
	shl	ecx, 3			# ecx = (4 - number of unaligned characters) * 8
					#     = 32 - number of unaligned bits
.if 1
	xor	eax, eax
	dec	eax			# eax = ~0
.else
	or	eax, -1			# eax = ~0
.endif
	shl	eax, cl			# eax = ~0 for unaligned characters, 0 elsewhere
	not	eax			# eax = 0 for unaligned characters, ~0 elsewhere
	or	eax, [edx]		# eax = unaligned characters
	jmp	.Ltwiddle
.Lnext:
	add	edx, 4			# edx = address of next 4 aligned characters
.Laligned:
	mov	eax, [edx]		# eax = next 4 aligned characters
.Ltwiddle:
	mov	ecx, eax
	sub	eax, 0x01010101
	not	ecx
	and	eax, 0x80808080
	and	ecx, eax		# eax = '\200' for matching characters, '\0' elsewhere
	jz	.Lnext			# no '\0' in any character?
.Lmatch:
	bsf	eax, ecx		# eax = offset of '\0' * 8 + 7
					#     = {7, 15, 23, 31}
	shr	eax, 3			# eax = offset of '\0'
					#     = {0, 1, 2, 3}
	add	eax, edx		# eax = address of '\0'
	sub	eax, [esp+4]		# eax = length of string
	ret

.size	strlen, .-strlen
.type	strlen, @function
.global	strlen
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strncat.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strncat.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = address of source
					# [esp+4] = address of destination
strncat:
	push	esi
	push	edi
	mov	esi, [esp+16]		# esi = address of source
	mov	edx, [esp+20]		# edx = count
	mov	edi, esi		# edi = address of source
	mov	ecx, edx		# ecx = count
	xor	eax, eax		# eax = '\0'
	repne	scasb
	sub	edx, ecx		# edx = length of source (including '\0')
	mov	edi, [esp+12]		# edi = address of destination
#	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	dec	edi			# edi = end of destination (address of '\0')
	mov	ecx, edx		# ecx = length of source (including '\0')
	rep	movsb
#	xor	eax, eax		# eax = '\0'
	stosb
	mov	eax, [esp+12]		# eax = address of destination
	pop	edi
	pop	esi
	ret

.size	strncat, .-strncat
.type	strncat, @function
.global	strncat
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strncmp.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strncmp.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = address of right string
					# [esp+4] = address of left string
strncmp:
	push	esi
	push	edi
	xor	eax, eax		# eax = 0
	mov	esi, [esp+12]		# esi = address of left string
	mov	edi, [esp+16]		# edi = address of right string
	cmp	edi, esi
	je	.Lequal			# address of right string = address of left string?

	mov	ecx, [esp+20]		# ecx = count
	test	ecx, ecx
	jz	.Lequal			# count = 0?

#	xor	eax, eax		# eax = 0,
#					# CF = 0, ZF = 1 (required when ecx is 0)
	repe	cmpsb
	seta	al			# eax = (left string > right string) ? 1
	sbb	eax, 0			# eax = (left string < right string) ? -1 : 0
					#     : (left string > right string) ? 1 : 0
					#     = {-1, 0, 1}
.Lequal:
	pop	edi
	pop	esi
	ret

.size	strncmp, .-strncmp
.type	strncmp, @function
.global	strncmp
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strncmp.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strncmp.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = address of right string
					# [esp+4] = address of left string
strncmp:
	push	edi
	push	esi
	mov	ecx, [esp+20]		# ecx = count
	test	ecx, ecx
	jz	.Lequal			# count = 0?

#	xor	eax, eax
	mov	edi, [esp+16]		# edi = address of right string
	mov	esi, [esp+12]		# esi = address of left string
#	cmp	esi, edi
#	je	.Lequal			# address of left string = address of right string?
.Lloop:
	lodsb
	scasb
	jne	.Lnotequal

	test	ax, ax
	jz	.Lequal

	dec	ecx
	jnz	.Lloop
.Lequal:
	xor	eax, eax
	pop	esi
	pop	edi
	ret
.Lnotequal:
	sbb	eax, eax		# eax = (left string < right string) ? -1 : 0
	or	eax, 1			# eax = (left string > right string) ? 1
					#     : (left string < right string) ? -1 : 0
					#     = {1, 0, -1}
	pop	esi
	pop	edi
	ret

.size	strncmp, .-strncmp
.type	strncmp, @function
.global	strncmp
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strncpy.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strncpy.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = address of source
					# [esp+4] = address of destination
strncpy:
	push	esi
	push	edi
	mov	esi, [esp+16]		# esi = address of source
	mov	edx, [esp+20]		# edx = count
	mov	edi, esi		# edi = address of source
	mov	ecx, edx		# ecx = count
	xor	eax, eax		# eax = '\0'
	repne	scasb
	sub	ecx, edx
	neg	ecx			# ecx = length of source (including '\0')
	sub	edx, ecx		# edx = count - length of source (including '\0')
	mov	edi, [esp+12]		# edi = address of destination
	rep	movsb
	mov	ecx, edx		# ecx = count - length of source (including '\0')
#	xor	eax, eax		# eax = '\0'
	rep	stosb
	mov	eax, [esp+12]		# eax = address of destination
	pop	edi
	pop	esi
	ret

.size	strncpy, .-strncpy
.type	strncpy, @function
.global	strncpy
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strnlen.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strnlen.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = count
					# [esp+4] = address of string
strnlen:
	mov	ecx, [esp+8]		# ecx = count
	test	ecx, ecx
	jz	.Lzero			# count = 0?

	xor	eax, eax		# eax = '\0'
	mov	edx, edi
	mov	edi, [esp+4]		# edi = address of string
#	test	edi, edi		# ZF = 0 (required when count is 0)
	repne	scasb			# ecx = (length of string < count)
					#     ? count - (length of string + 1) : 0
	neg	ecx			# CF = (ecx <> 0)
					#    = ([edi] = '\0')
					#    = (length of string < count),
					# ecx = (length of string < count)
					#     ? length of string + 1 - count : 0
	sbb	ecx, eax		# ecx = (length of string < count)
					#     ? length of string - count : 0
	add	ecx, [esp+8]		# ecx = (length of string < count)
					#     ? length of string : count
	mov	edi, edx
.Lzero:
	mov	eax, ecx		# eax = (length of string < count)
					#     ? length of string : count
	ret

.size	strnlen, .-strnlen
.type	strnlen, @function
.global	strnlen
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strnset.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strnset.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = character
					# [esp+4] = address of string
strnset:
	mov	edx, [esp+4]		# edx = address of string
	mov	ecx, [esp+12]		# ecx = count
	test	ecx, ecx
	jz	.Ldone			# count = 0?

	xor	eax, eax		# eax = '\0'
	push	edi
	mov	edi, edx		# edi = address of string
#	test	edi, edi		# ZF = 0 (required when count is 0)
	repne	scasb			# ecx = (length of string < count)
					#     ? count - (length of string + 1) : 0
	mov	edi, edx		# edi = address of string
	neg	ecx			# CF = (ecx <> 0)
					#    = ([edi] = '\0')
					#    = (length of string < count),
					# ecx = (length of string < count)
					#     ? length of string + 1 - count : 0
	sbb	ecx, eax		# ecx = (length of string < count)
					#     ? length of string - count : 0
	add	ecx, [esp+16]		# ecx = (length of string < count)
					#     ? length of string : count
	mov	eax, [esp+12]		# eax = character
	stosb
	pop	edi
.Ldone:
	mov	eax, edx		# eax = address of string
	ret

.size	strnset, .-strnset
.type	strnset, @function
.global	strnset
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strpbrk.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strpbrk.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = address of delimiter
					# [esp+4] = address of string
strpbrk:
	mov	eax, [esp+4]		# eax = address of string
	mov	edx, [esp+8]		# edx = address of delimiter
	xor	ecx, ecx		# ecx = 0
	cmp	cl, [edx]
	je	.Lempty			# delimiter[0] = '\0'?

	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx			# bitmap[0..255] = 0,
					# esp = address of bitmap
.Lsetup:
	bts	[esp], ecx		# bitmap[ecx] = 1
	mov	cl, [edx]		# ecx = *delimiter
	inc	edx			# edx = ++delimiter
	cmp	cl, ch
	jne	.Lsetup			# ecx <> '\0'?
.Lskip:
	mov	cl, [eax]		# ecx = *string
	inc	eax			# eax = ++string
	bt	[esp], ecx
	jnc	.Lskip			# bitmap[ecx] = 0 (no match)?
.Lstop:
	dec	eax			# eax = --string
	neg	ecx
	sbb	ecx, ecx		# ecx = (*string = '\0') ? 0 : ~0
	and	eax, ecx		# eax = (*string = '\0') ? 0 : address of string
	add	esp, 32
	ret
.Lempty:
	xor	eax, eax
	ret

.size	strpbrk, .-strpbrk
.type	strpbrk, @function
.global	strpbrk
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strrchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strrchr.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = character
					# [esp+4] = address of string
strrchr:
	mov	edx, edi
	mov	edi, [esp+4]		# edi = address of string
	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	not	ecx			# ecx = length of string (including '\0')
	dec	edi			# edi = end of string (address of '\0')
	mov	eax, [esp+8]		# eax = character
	std
	repne	scasb
	cld
	inc	edi			# edi = address of character
	neg	ecx			# CF = (ecx <> 0)
					#    = ([edi] = character)
	sbb	eax, eax		# eax = (ecx = 0) ? 0 : -1
	and	eax, edi		# eax = (ecx = 0) ? 0 : address of character
	mov	edi, edx
	ret

.size	strrchr, .-strrchr
.type	strrchr, @function
.global	strrchr
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strrev.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strrev.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+4] = address of string
strrev:
	mov	edx, edi
	mov	edi, [esp+4]		# edi = string
	xor	eax, eax
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	add	ecx, edi		# ecx = address of string - 1
	dec	edi			# edi = address of '\0'
	jmp	.Lcontinue
.Lreverse:
	mov	al, [ecx]
	mov	ah, [edi]
	mov	[ecx], ah
	mov	[edi], al
.Lcontinue:
	inc	ecx
	dec	edi
	cmp	edi, ecx
	ja	.Lreverse

	mov	edi, edx
	mov	eax, edx		# eax = address of string
	ret

.size	strrev, .-strrev
.type	strrev, @function
.global	strrev
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strset.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strset.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = character
					# [esp+4] = address of string
strset:
	mov	edx, edi
	mov	edi, [esp+4]		# edi = address of string
	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	not	ecx			# ecx = length of string (including '\0')
	sub	edi, ecx		# edi = address of string
	dec	ecx			# ecx = length of string
	mov	eax, [esp+8]		# eax = character
	rep	stosb
	mov	edi, edx
	mov	eax, [esp+4]		# eax = address of string
	ret

.size	strset, .-strset
.type	strset, @function
.global	strset
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strspn.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strspn.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = address of delimiter
					# [esp+4] = address of string
strspn:
	mov	eax, [esp+4]		# eax = address of string
	mov	edx, [esp+8]		# edx = address of delimiter
	xor	ecx, ecx		# ecx = 0
	cmp	cl, [edx]
	je	.Lempty			# delimiter[0] = '\0'?

	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx
	push	ecx			# bitmap[0..255] = 0,
					# esp = address of bitmap
	mov	cl, [edx]		# ecx = *delimiter
	inc	edx			# edx = ++delimiter
.Lsetup:
	bts	[esp], ecx		# bitmap[ecx] = 1
	mov	cl, [edx]		# ecx = *delimiter
	inc	edx			# edx = ++delimiter
	cmp	cl, ch
	jne	.Lsetup			# ecx <> '\0'?

	mov	edx, eax		# edx = address of string
.Lskip:
	mov	cl, [eax]		# ecx = *string
	inc	eax			# eax = ++string
	bt	[esp], ecx
	jc	.Lskip			# bitmap[ecx] = 1 (match)?
.if 1
	dec	eax			# eax = --string
	sub	eax, edx
.else
	stc
	sbb	eax, edx		# eax = count of matching characters
.endif
	add	esp, 32
	ret
.Lempty:
	xor	eax, eax
	ret

.size	strspn, .-strspn
.type	strspn, @function
.global	strspn
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strstr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strstr.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+8] = address of needle
					# [esp+4] = address of haystack
strstr:
	push	edi
	mov	edi, [esp+12]		# edi = address of needle
	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	not	ecx			# ecx = length of needle (including '\0')
	dec	ecx			# ecx = length of needle
	mov	eax, [esp+8]		# eax = address of haystack
	jz	.Lempty			# length of needle = 0?

	mov	edx, ecx		# edx = length of needle
.ifdef SIMPLE
	push	esi
.Lcompare:
	mov	esi, eax		# esi = current address in haystack
	mov	edi, [esp+16]		# edi = address of needle
	mov	ecx, edx		# ecx = length of needle
	repe	cmpsb
	je	.Lmatch			# needle in haystack?

	inc	eax			# eax = next address in haystack
	cmp	byte ptr [esi-1], 0
	jne	.Lcompare		# non-matching character in haystack <> '\0'?

	xor	eax, eax
.Lmatch:
.else # SIMPLE
	mov	edi, eax		# edi = address of haystack
	xor	eax, eax		# eax = '\0'
	xor	ecx, ecx
	dec	ecx			# ecx = -1
	repne	scasb
	not	ecx			# ecx = length of haystack (including '\0')
	sub	edi, ecx		# edi = address of haystack
	dec	ecx			# ecx = length of haystack
	jz	.Lempty			# length of haystack = 0?

	cmp	ecx, edx
	jb	.Lempty			# length of haystack < length of needle?

	push	esi
	push	ebx
.Lsearch:
	mov	esi, [esp+20]		# esi = address of needle
	mov	al, [esi]		# al = first character of needle
					# edi = current address in haystack
	repne	scasb			# edi = next address in haystack,
					# ecx = current length of haystack
	jne	.Lbreak			# (first character of) needle not in haystack?

	dec	ecx			# ecx = next length of haystack

	mov	al, [esi+edx-1]		# al = last character of needle
	cmp	al, [edi+edx-2]
	jne	.Lcontinue		# last character of needle not in haystack?
.Lcompare:
	mov	eax, edi		# eax = next address in haystack
	mov	ebx, ecx		# ebx = next length of haystack
.if 0
	dec	edi			# edi = address of matching character
					#     = current address in haystack
					# esi = address of needle
	mov	ecx, edx		# ecx = length of needle
.else
					# edi = next address in haystack
	inc	esi			# esi = address of needle + 1
	mov	ecx, edx
	dec	ecx			# ecx = length of needle - 1
#	jz	.Lmatch			# needle in haystack?
.endif
	repe	cmpsb
	je	.Lmatch			# needle in haystack?

	mov	edi, eax		# edi = current address in haystack
	mov	ecx, ebx		# ecx = current length of haystack
.Lcontinue:
	cmp	ecx, edx
	jae	.Lsearch		# length of haystack >= length of needle?
.Lbreak:
	xor	eax, eax
	pop	ebx
	pop	esi
	pop	edi
	ret
.Lmatch:
	dec	eax			# eax = address of needle in haystack
	pop	ebx
.endif # SIMPLE
	pop	esi
.Lempty:
	pop	edi
	ret

.size	strstr, .-strstr
.type	strstr, @function
.global	strstr
.end
+EOF

as --32 --no-pad-sections --noexecstack -o strtok_r.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"strtok_r.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = address of address of next
					# [esp+8] = address of delimiter
					# [esp+4] = address of string
strtok_r:
	mov	ecx, [esp+4]		# ecx = address of string
	mov	eax, [esp+8]		# eax = address of delimiter
	mov	edx, [esp+12]		# edx = address of address of next
	test	ecx, ecx
	jnz	.Lstart			# address of string <> 0?

	or	ecx, [edx]		# ecx = address of next
	jz	.Lnull			# address of next = 0 = address of string?
.Lstart:
	cmp	byte ptr [ecx], 0
	je	.Lnull			# string[0] = '\0'?

	cmp	byte ptr [eax], 0
	je	.Lempty			# delimiter[0] = '\0'?

	push	ebx
	xor	ebx, ebx		# ebx = 0
	push	ebx
	push	ebx
	push	ebx
	push	ebx
	push	ebx
	push	ebx
	push	ebx
	push	ebx			# bitmap[0..255] = 0,
					# esp = address of bitmap
	mov	bl, [eax]		# ebx = *delimiter
	inc	eax			# eax = ++delimiter
.Lsetup:
	bts	[esp], ebx		# bitmap[ebx] = 1
	mov	bl, [eax]		# ebx = *delimiter
	inc	eax			# eax = ++delimiter
	cmp	bl, bh
	jne	.Lsetup			# ebx <> '\0'?
.Lskip:
	mov	bl, [ecx]		# ebx = *string
	inc	ecx			# ecx = ++string
	bt	[esp], ebx
	jc	.Lskip			# bitmap[ebx] = 1?
					# (ebx is a delimiter)?
	cmp	bl, bh
	je	.Lnone			# ebx = '\0'?

	mov	bl, bh			# ebx = 0
	bts	[esp], ebx		# bitmap['\0'] = 1
	mov	eax, ecx
	dec	eax			# eax = address of token
.Ltoken:
	mov	bl, [ecx]		# ebx = *string
	inc	ecx			# ecx = ++string
	bt	[esp], ebx
	jnc	.Ltoken			# bitmap[ebx] = 0?
					# (ebx is not a delimiter)?
	cmp	bl, bh
	je	.Llast			# ebx = '\0'?

	mov	[ecx-1], bh		# string[-1] = '\0' (terminate token)
	mov	[edx], ecx		# *next = address of string
	add	esp, 32
	pop	ebx
	ret
.Lnone:
	mov	eax, ebx		# eax = 0
.Llast:
	mov	[edx], ebx		# *next = 0
	add	esp, 32
	pop	ebx
	ret
.Lnull:
	xor	eax, eax		# eax = 0
	mov	[edx], eax		# *next = 0
	ret
.Lempty:
	mov	eax, ecx		# eax = address of string
	xor	ecx, ecx
	mov	[edx], ecx		# *next = 0
	ret

.size	strtok_r, .-strtok_r
.type	strtok_r, @function
.global	strtok_r
.end
+EOF

ar rsv string-i386.a string.o strcat.o strchr.o strcmp.o strcoll.o strcpy.o strcspn.o strlen.o strncat.o strncmp.o strncpy.o strnlen.o strpbrk.o strrchr.o strrev.o strset.o strspn.o strstr.o strtok_r.o
rm string.o strcat.o strchr.o strcmp.o strcoll.o strcpy.o strcspn.o strlen.o strncat.o strncmp.o strncpy.o strnlen.o strpbrk.o strrchr.o strrev.o strset.o strspn.o strstr.o strtok_r.o
else
echo "Usage: $0 { amd64 | i386 }"
echo
echo "Create object library 'string-amd64.a' or 'string-i386.a' with properly"
echo "optimised string routines for Clang and GCC, as (partial) replacement for"
echo "either 'libclang-rt.builtins-x86_64.a' or 'libclang-rt.builtins-i386.a' as"
echo "well as 'libgcc.a'"
fi

if [ $# -eq 1 -a "$1" == 'amd64' ]
then
gcc -m64 -x c - -x none string-amd64.a
elif [ $# -eq 1 -a "$1" = 'i386' ]
then
gcc -m32 -x c - -x none string-i386.a
else
false
fi <<+\E\O\F && ./a.out
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

#include <stdio.h>
#include <string.h>

#ifdef ALTERNATE
char	*strpbrk(char const *string, char const *delimiter)
{
	string += strcspn(string, delimiter);

	return *string == '\0' ? NULL : (char *) string;
}

char	*strtok_r(char *string, char const *delimiter, char **next)
{
	if (string == NULL)
		string = *next;

	if ((string == NULL) || (*string == '\0'))
		return *next = NULL;

	string += strspn(string, delimiter);

	if (*string == '\0')
		return *next = NULL;

	*next = string + strcspn(string, delimiter);

	if (**next != '\0')
		*(*next)++ = '\0';
	else
		*next = NULL;

	return string;
}
#endif

int	main(void)
{
	char	*token, *ptr0, *ptr1;
	static	char	str0[] = "* * *";
	static	char	str1[] = "?abc???#$@,,,#xyz";
	int	error = 0;

	printf("test 0: strtok_r(%p, \" *\", %p) = ", str0, ptr0);
	token = strtok_r(str0, " *", &ptr0);	// null pointer
	printf("%p\t\"%s\"\n", token, token);
	error += token != NULL;

	printf("test 1: strtok_r(%p, \"?\", %p) = ", str1, ptr1);
	token = strtok_r(str1, "?", &ptr1);	// pointer to token "abc"
	printf("%p\t\"%s\"\n", token, token);
	error += strcmp(token, "abc") != 0;

	printf("test 2: strtok_r(NULL, \",\", %p) = ", ptr1);
	token = strtok_r(NULL, ",", &ptr1);	// pointer to token "??#$@"
	printf("%p\t\"%s\"\n", token, token);
	error += strcmp(token, "??#$@") != 0;

	printf("test 3: strtok_r(NULL, \"#,\", %p) = ", ptr1);
	token = strtok_r(NULL, ",#", &ptr1);	// pointer to token "xyz"
	printf("%p\t\"%s\"\n", token, token);
	error += strcmp(token, "xyz") != 0;

	printf("test 4: strtok_r(NULL, \"?\", %p) = ", ptr1);
	token = strtok_r(NULL, "?", &ptr1);	// null pointer
	printf("%p\t\"%s\"\n", token, token);
	error += token != NULL;

	if (error == 0)
		printf("strtok_r() passed all tests\n");
	else
		printf("strtok_r() FAILED %u tests\n", error);

	return error;
}
+EOF
